# $Id: vCard.pm,v 1.12 2004/10/05 12:53:59 asc Exp $
use strict;

package XML::Generator::vCard;
use base qw (XML::SAX::Base);

$XML::Generator::vCard::VERSION = '1.0_02';

=head1 NAME

XML::Generator::vCard - generate SAX2 events for vCard 3.0

=head1 SYNOPSIS

 use XML::SAX::Writer;
 use XML::Generator::vCard;

 my $writer = XML::SAX::Writer->new();
 my $driver = XML::Generator::vCard->new(Handler=>$writer);

 $driver->parse_files("test.vcf");

=head1 DESCRIPTION

Generate SAX2 events for vCard 3.0

This package supersedes I<XML::SAXDriver::vCard>.

=cut

use Encode;
use MIME::Base64;
use Text::vCard::Addressbook;

# This on died in IETF draft land and was
# subsequently sucked into the vortex. I
# would like nothing more than some definitive
# vcard-xml thingy that I could point (please,
# don't say RDF; not yet anyway) so in the
# interim this will have to do...

# http://www.ietf.org/internet-drafts/ \
#  draft-dawson-vCard-xml-dtd-04.txt

use constant NS_VCARD => "http://aaronland.info/perl/xml/generator/vcard#";
use constant NS_FOAF  => "http://xmlns.com/foaf/0.1/";

use constant VCARD_VERSION => "3.0";

=head1 PACKAGE METHODS

=cut

=head2 __PACKAGE__->new(%args)

This method inherits from I<XML::SAX::Base>.

Returns a I<XML::Generator::vCard> object.

=cut

=head1 OBJECT METHODS

=cut

=head2 $pkg->parse_files(@files)

=cut

sub parse_files {
  my $self  = shift;
  my @files = @_;

  my $book = ();

  eval {
      $book = Text::vCard::Addressbook->load(\@files);
  };

  if ($@) {
      warn $@;
      return 0;
  }

  return $self->_render_doc([ $book->vcards() ]);
}

=head1 PRIVATE METHODS

Private methods are documented below in case you need to subclass
this package to tweak its output.

=cut

=head2 $obj->_render_doc(\@vcards)

=cut

sub _render_doc {
    my $self  = shift;
    my $cards = shift;

    $self->_start_document();

    if (scalar(@$cards) > 1) {
	
	$self->start_element({Name=>"vCardSet"});
	
	foreach my $vcard (@$cards) {
	    $self->_render_card($vcard);
	}
	
	$self->end_element({Name=>"vCardSet"});
    }
    
    else {
	$self->_render_card($cards->[0]);
    }
    
    #
    
    $self->_end_document();
    return 1;
}

=head2 $obj->_render_card(Text::vCard)

=cut

sub _render_card {
  my $self  = shift;
  my $vcard = shift;
  
  my $attrs = {
      "{}version" => {Name  => "version",
		      Value => ($vcard->version() || VCARD_VERSION)},
      "{}class"   => {Name  => "class",
		      Value => ($vcard->class() || "PUBLIC")},
  };

  #

  foreach my $prop ("uid","rev","prodid") {
      if (my $value = $vcard->$prop()) {
	  $attrs->{"{}$prop"} = {Name  => $prop,
				 Value => $value};
      }
  }

  $self->start_element({Name       => "vCard",
			Attributes => $attrs});

  # 

  $self->_render_fn($vcard);
  $self->_render_n($vcard);
  $self->_render_nickname($vcard);
  $self->_render_photo($vcard);
  $self->_render_bday($vcard);
  $self->_render_adrs($vcard);
  $self->_render_labels($vcard);
  $self->_render_tels($vcard);
  $self->_render_emails($vcard);
  $self->_render_instantmessaging($vcard);
  $self->_render_mailer($vcard);
  $self->_render_tz($vcard);
  $self->_render_geo($vcard);
  $self->_render_org($vcard);
  $self->_render_title($vcard);
  $self->_render_role($vcard);
  $self->_render_logo($vcard);
  # AGENT
  $self->_render_categories($vcard);
  $self->_render_note($vcard);
  # SORT
  $self->_render_sound($vcard);
  $self->_render_url($vcard);
  $self->_render_key($vcard);
  $self->_render_custom($vcard);

  $self->end_element({Name=>"vCard"});

  return 1;
}

=head2 $obj->_render_fn(Text::vCard)

=cut

sub _render_fn {
    my $self = shift;
    my $vcard = shift;

    $self->_pcdata({name  => "fn",
		    value => $vcard->fn()});
    
    return 1;
}

=head2 $obj->_render_n(Text::vCard)

=cut

sub _render_n {
    my $self  = shift;
    my $vcard = shift;

    my $n = $vcard->get({'element_type'=>'name'});

    if (! $n) {
	return 1;
    }

    $n = $n->[0];

    #

    if (($n->family()) || ($n->given())) {
	$self->start_element({Name=>"n"});
	
	$self->_pcdata({name  => "family",
			value => $n->family()});
	
	    $self->_pcdata({name  => "given",
			    value => $n->given()});
	
	if (my $o = $n->middle()) {
	    $self->_pcdata({name  => "other",
			    value => $o});
	}
	
	if (my $p = $n->prefixes()) {
	    $self->_pcdata({name  => "prefix",
			    value => $p});
	}
	
	if (my $s = $n->suffixes()) {
	    $self->_pcdata({name  => "suffix",
			    value => $s});
	}
	
	$self->end_element({Name=>"n"});
    }     
    
    return 1;
}

=head2 $obj->_render_nickname(Text::vCard)

=cut

sub _render_nickname {
    my $self  = shift;
    my $vcard = shift;

    if (my $nick = $vcard->nickname()) {
	$self->_pcdata({name  => "nickname",
			value => $nick});
    }

    return 1;
}

=head2 $obj->_render_photo(Text::vCard)

=cut

sub _render_photo {
    my $self  = shift;
    my $vcard = shift;

    my $photos = $vcard->get({"element_type" => "photo"});

    if (! $photos) {
	return 1;
    }

    foreach my $p (@$photos) {
	$self->_media({name   => "photo",
		       media  => $p});
    }
    
    return 1;
}


=head2 $obj->_render_bday(Text::vCard)

=cut

sub _render_bday {
    my $self = shift;
    my $vcard = shift;

    if (my $bday = $vcard->bday()) {
	$self->_pcdata({name  => "bday",
			value => $bday});
    }

    return 1;
}

=head2 $obj->_render_adrs(Text::vCard)

=cut

sub _render_adrs {
    my $self  = shift;
    my $vcard = shift;

    my $addresses = $vcard->get({"element_type" => "addresses"});

    if (! $addresses) {
	return 1;
    }

    #

    foreach my $adr (@$addresses) {
		
	my $types = join(";",$adr->types());
	
	$self->start_element({Name       => "adr",
			      Attributes => {"{}del.type"=>{Name  => "del.type",
							    Value => $types}}
			  });
	
	if (my $p = $adr->po_box()) {
	    $self->_pcdata({name  => "pobox",
			    value => $p});
	}
	
	if (my $e = $adr->extended()) {
	    $self->_pcdata({name  => "extadr",
			    value => $e});
	}
	
	if (my $s = $adr->street()) {
	    $self->_pcdata({name  => "street",
			    value => $s});
	}
	
	if (my $c = $adr->city()) {
	    $self->_pcdata({name  => "locality",
			    value => $c});
	}
	
	if (my $r = $adr->region()) {
	    $self->_pcdata({name  => "region",
			    value => $r});
	}
	
	if (my $p = $adr->post_code()) {
	    $self->_pcdata({name  => "pcode",
			    value => $p});
	}
	
	if (my $c = $adr->country()) {
	    $self->_pcdata({name  => "country",
			    value => $c});
	}
	
	$self->end_element({Name=>"adr"});
    }
    
    return 1;
}


=head2 $obj->_render_labels(Text::vCard)

=cut

sub _render_labels {
    my $self  = shift;
    my $vcard = shift;

    my $labels = $vcard->get({'element_type' => 'labels'});
    
    if (! $labels) {
	return 1;
    }

    #

    foreach my $l (@$labels) {
	
	my $types = join(";",$l->types());
	
	$self->_pcdata({name  => "label",
			value => $l->value(),
			attrs => {"{}del.type"=>{Name  => "del.type",
						 Value => $types}}
		    });
    }
    
    return 1;
}

=head2 $obj->_render_tels(Text::vCard)

=cut

sub _render_tels {
    my $self  = shift;
    my $vcard = shift;

    my $numbers = $vcard->get({'element_type' => 'phone'});

    if (! $numbers) {
	return 1;
    }

    #

    foreach my $tel (@$numbers) {
	
	my $types = join(";",$tel->types());
	
	$self->_pcdata({name  => "tel",
			value => $tel->value(),
			attrs => {"{}tel.type"=>{Name  => "tel.type",
						 Value => $types}}
		    });
    }
    
    return 1;
}

=head2 $obj->_render_emails(Text::vCard)

=cut

sub _render_emails {
    my $self  = shift;
    my $vcard = shift;

    my $addresses = $vcard->get({'element_type' => 'email'});

    if (! $addresses) {
	return 1;
    }

    #

    foreach my $e (@$addresses) {

	my $types = join(";",$e->types());
	
	$self->_pcdata({name  => "email",
			value => $e->value(),
			attrs => {"{}email.type"=>{Name  => "email.type",
						   Value => $types}}
		    });
    }
    
    return 1;
}

=head2 $obj->_render_instantmessaging(Text::vCard)

=cut

sub _render_instantmessaging {
    my $self  = shift;
    my $vcard = shift;

    my $im_list = $self->_im_services();

    foreach my $service (sort {$a cmp $b} keys %$im_list) {

	my $addresses = $vcard->get({'element_type' => "x-$service"});
	
	if (! $addresses) {
	    next;
	}

	foreach my $im (@$addresses) {
	
	    my $types = join(";",$im->types());
	
	    $self->_pcdata({name  => $im_list->{$service},
			    value => $im->value(),
			    attrs => {"{}im.type"=> {Name  => "im.type",
						     Value => $types}}
			});
	}
    }

    return 1;
}

=head2 $obj->_render_mailer(Text::vCard)

=cut

sub _render_mailer {
    my $self  = shift;
    my $vcard = shift;

    if (my $m = $vcard->mailer()) {
	$self->_pcdata({name  => "mailer",
			value => $m});
    }

    return 1;
}

=head2 $obj->_render_tz(Text::vCard)

=cut

sub _render_tz {
    my $self  = shift;
    my $vcard = shift;

    if (my $tz = $vcard->tz()) {
	$self->_pcdata({name  => "tz",
			value => $tz});
    }

    return 1;
}

=head2 $obj->_render_geo(Text::vCard)

=cut

sub _render_geo {
    my $self  = shift;
    my $vcard = shift;

    my $geo = $vcard->get({'element_type' => "geo"});

    if (! $geo) {
	return 1;
    }

    $geo = $geo->[0];

    #

    $self->start_element({Name=>"geo"});

    $self->_pcdata({name  => "lat",
		    value => $geo->lat()});

    $self->_pcdata({name  => "lon",
		    value => $geo->long()});

    $self->end_element({Name=>"geo"});
    
    return 1;
}

=head2 $obj->_render_org(Text::vCard)

=cut

sub _render_org {
    my $self = shift;
    my $vcard = shift;

    my $orgs = $vcard->get({'element_type' => "org"});

    if (! $orgs) {
	return 1;
    }

    #

    foreach my $o (@$orgs) {
	$self->start_element({Name=>"org"});

	if (my $name = $o->name()) {
	    $self->_pcdata({name  => "orgnam",
			    value => $name});
	} 

	if (my $unit = $o->unit()) {
	    $self->_pcdata({name  => "orgunit",
			    value => $unit});
	}

	$self->end_element({Name=>"org"});
    }

    return 1;
}

=head2 $obj->_render_title(Text::vCard)

=cut

sub _render_title {
    my $self  = shift;
    my $vcard = shift;

    if (my $t = $vcard->title()) {
	$self->_pcdata({name  => "title",
			value => $t});
    }

    return 1;
}

=head2 $obj->_render_role(Text::vCard)

=cut

sub _render_role {
    my $self = shift;
    my $vcard = shift;

    if (my $r = $vcard->role()) {
	$self->_pcdata({name  => "role",
			value => $r});
    }

    return 1;
}

=head2 $obj->_render_logo(Text::vCard)

=cut

sub _render_logo {
    my $self  = shift;
    my $vcard = shift;

    my $logos = $vcard->get({'element_type' => 'logo'});

    if (! $logos) {
	return 1;
    }

    foreach my $l (@$logos) {
	$self->_media({name  => "logo",
		       media => $l});
    }

    return 1;
}

=head2 $obj->_render_categories(Text::vCard)

=cut

sub _render_categories {
    my $self = shift;
    my $vcard = shift;

    my $cats = $vcard->get({'element_type' => 'categories'}) ||
	       $vcard->get({'element_type' => 'category'});

    if (! $cats) {
	return 1;
    }

    foreach my $c (@$cats) {
	
	$self->start_element({Name=>"categories"});
	
	foreach (split(",",$c->value())) {
	    $self->_pcdata({name  => "item",
			    value => $_});
	}
	
	$self->end_element({Name=>"categories"});
    }

    return 1;
}

=head2 $obj->_render_note(Text::vCard)

=cut

sub _render_note {
    my $self = shift;
    my $vcard = shift;

    my $n = $vcard->get({'element_type'=>'note'});

    if (! $n) {
	return 1;
    }

    if (my $n = $vcard->note()) {
	$self->_pcdata({name  => "note",
			cdata => 1,
			value => $n});	
    }
    
    return 1;
}

=head2 $self->_render_sound(Text::vCard)

=cut

sub _render_sound {
    my $self = shift;
    my $vcard = shift;

    my $snds = $vcard->get({'element_type' => 'sound'});

    if (! $snds) {
	return 1;
    }

    foreach my $s (@$snds) {
	$self->_media({name  => "sound",
		       media => $s});
    }

    return 1;
}

=head2 $self->_render_url(Text::vCard)

=cut

sub _render_url {
    my $self  = shift;
    my $vcard = shift;

    if (my $url = $vcard->url()) {
	$self->_pcdata({name  => "url",
			attrs => {"{}uri" => {Name  => "uri",
						   Value => $url}}});
    }
    
    return 1;
}

=head2 $obj->_render_key(Text::vCard)

=cut

sub _render_key {
    my $self  = shift;
    my $vcard = shift;

    my $keys = $vcard->get({'element_type' => 'key'});

    if (! $keys) {
	return 1;
    }

    foreach my $k (@$keys) {
	$self->_media({name  => "key",
		       media => $k});
    }

    return 1;
}

=head2 $obj->_render_custom(Text::vCard)

By default this method does nothing. It is here to
be subclassed.

=cut

sub _render_custom {
    my $self  = shift;
    my $vcard = shift;

    return 1;
}

=head2 $obj->_im_services()

Returns a hash ref mapping an instant messaging service
type to an XML element. Default is :

 {"aim"    => "foaf:aimChatID",
  "yahoo"  => "foaf:yahooChatID",
  "msn"    => "foaf:msnChatID",
  "jabber" => "foaf:JabberID",
  "icq"    => "foaf:icqChatId"}

This is called by the I<_render_instantmessaging> method.

=cut

sub _im_services {
    return {"aim"    => "foaf:aimChatID",
	    "yahoo"  => "foaf:yahooChatID",
	    "msn"    => "foaf:msnChatID",
	    "jabber" => "foaf:JabberID",
	    "icq"    => "foaf:"};
}

sub _pcdata {
  my $self = shift;
  my $data = shift;
  $self->start_element({Name=>$data->{name},Attributes=>$data->{attrs}});
  $self->start_cdata() if ($data->{cdata});
  $self->characters({Data=>encode_utf8($data->{value})});
  $self->end_cdata() if ($data->{cdata});
  $self->end_element({Name=>$data->{name}});
  return 1;
}

sub _media {
  my $self = shift;
  my $data = shift;

  my $attrs = {};

  # as in not 'key' and not something pointing to an 'uri'

  if (($data->{name} !~ /^k/) && ($data->{type})) {

      # as in 'photo' or 'logo' 
      # and not 'sound'
      
      my $mime = ($data->{name} =~ /^[pl]/i) ? "img" : "aud";
      
      $attrs = {"{}$mime.type"=>{Name  => "$mime.type",
				 Value => $data->{type}}};
  }

  #

  my $obj = $data->{media};

  $self->start_element({Name       => $data->{name},
			Attributes => $attrs});

  if ($obj->is_type("base64")) {
      $self->_pcdata({name  => "b64bin",
		      value => encode_base64($obj->value()),
		      cdata => 1});
  }

  else {
      $self->_pcdata({name  => "extref",
		      attrs => {"{}uri" => {Name  => "uri",
					    Value => $obj->value()}}
		  });
  }

  $self->end_element({Name=>$data->{name}});
  return 1;
}

sub _start_document {
    my $self = shift;

    $self->start_document();
    $self->xml_decl({Version  => "1.0",
		     Encoding => "UTF-8"});
    
    $self->_set_prefix_mapping();
    return 1;
}

sub _end_document {
    my $self = shift;

    $self->_unset_prefix_mapping();
    $self->end_document();
    return 1;
}

sub _set_prefix_mapping {
    my $self = shift;

    $self->start_prefix_mapping({Prefix       => "",
				 NamespaceURI => NS_VCARD});

    $self->start_prefix_mapping({Prefix       => "foaf",
				 NamespaceURI => NS_FOAF});

    return 1;
}

sub _unset_prefix_mapping {
    my $self = shift;

    $self->end_prefix_mapping({Prefix => "foaf"});
    $self->end_prefix_mapping({Prefix => ""});

    return 1;
}

sub DESTROY {}

=head1 HOW TO

=head2 Generate SAX events for a custom 'X-*' field

 package MyGenerator;
 use base qw (XML::Generator::vCard);

 sub _render_custom {
   my $self  = shift;
   my $vcard = shift;

   my $custom = $vcard->get({"element_type" => "x-foobar"});
	
   if (! $addresses) {
      next;
   }

   foreach my $foo (@$custom) {
	
      my $types = join(";",$foo->types());
	
      $self->_pcdata({name  => "foo:bar",
	   	      value => $foo->value(),
		      attrs => {"{}type"=> {Name  => "type",
					    Value => $types}}
		      });
   }

   return 1;
 }
 
 package main;

 my $writer = XML::SAX::Writer->new();
 my $parser = MyGenerator->new(Handler=>$writer);

 $parser->parse_files(@ARGV);

=head2 Add custom namespaces

 package MyGenerator;
 use base qw (XML::Generator::vCard);

 sub _set_prefix_mapping {
    my $self = shift;

    $self->start_prefix_mapping({Prefix       => "foo",
				 NamespaceURI => "x-urn:foo:bar:});

    return $self->SUPER::_set_prefix_mapping();
 }

 sub _unset_prefix_mapping {
    my $self = shift;

    $self->end_prefix_mapping({Prefix => "foo"});
    return $self->SUPER::_unset_prefix_mapping();
 }

 package main;

 my $writer = XML::SAX::Writer->new();
 my $parser = MyGenerator->new(Handler=>$writer);

 $parser->parse_files(@ARGV);

=cut

=head1 VERSION

1.0_02

=head1 DATE

$Date: 2004/10/05 12:53:59 $

=head1 AUTHOR

Aaron Straup Cope E<lt>ascope@cpan.orgE<gt>

=head1 SEE ALSO

L<Text::vCard>

http://www.ietf.org/rfc/rfc2426.txt

http://www.ietf.org/rfc/rfc2425.txt

http://www.globecom.net/ietf/draft/draft-dawson-vcard-xml-dtd-03.html

=head1 BUGS

Please report all bugs via http://rt.cpan.org

=head1 LICENSE

Copyright (c) 2004, Aaron Straup Cope. All Rights Reserved.

This is free software, you may use it and distribute it 
under the same terms as Perl itself.

=cut

return 1;
